(** This file is based on Leory's contribution to POPLmark challenge
  and is modified for STLC (Simply typed lambda calculus). And it
  establishes the soundness of the Church-style simply-typed lambda
  calculus (STLC). 

  We tried to use the same definitions, notations, and structuresa as
  much as possible. Therefore, many of Leroy's comments remain the
  same or slightly modified. Our own new comments are added if
  necessary. 

  The main differences from Leroy's orginal work are as follows:

  - Use of simultaneous substitutions and renaming.
  - No use of swapping technique
  - Pencil-and-paper style of binding *)

Set Implicit Arguments.

Require Import Arith.
Require Import List.

Require Import extralibrary.

Require Import sublist.
Require Import environments.
Require Import environments_extra.
Require Import Decidable.

(** * Names *)

(** We use names to represent variables in terms. Any infinite type
  with decidable equality will do. *) 

Inductive name_kind : Set :=
  | TERM: name_kind.

Definition name : Set := (name_kind * nat)%type.

Definition kind (n: name) : name_kind := fst n.

(** Equality between names is decidable. *)

Lemma eq_name: forall (n1 n2: name), {n1 = n2} + {n1 <> n2}.
Proof.
  assert (forall k1 k2: name_kind, {k1 = k2} + {k1 <> k2});
  decide equality.
  assert (forall n1 n2: nat, {n1 = n2} + {n1 <> n2}); 
  decide equality.
Defined.

(** The EqDec class is a operational class such that we can use 
  notation for all of its instances. *)

Instance name_EqDec : EqDec name :=
  {
    dec := eq_name
  }.

(** Moreover, we have the following obvious simplification rules on
  tests over name equality. *)

Lemma eq_name_true {A: Set} (n: name) (a b: A) :
  (if n == n then a else b) = a.
Proof.
  intros. case_var; congruence.
Qed.

Lemma eq_name_false {A: Set} (n m: name) (a b: A) :
  n <> m -> (if n == m then a else b) = b.
Proof.
  intros. case_var; congruence.
Qed.

(** The following lemma shows that there always exists a name of the
  given kind that is fresh w.r.t. the given list of names, that is,
  distinct from all the names in this list. *) 

Definition find_fresh_name (k: name_kind) (l: list name) : name :=
  (k, 1 + fold_right (fun (n:name) x => max (snd n) x) 0 l)%nat.

Lemma find_fresh_name_is_fresh:
  forall k l, let n := find_fresh_name k l in ~In n l /\ kind n = k.
Proof.
  intros.
  set (ident := fun (n: name) => snd n). 
  set (maxid := 
   fold_right (fun (n:name) x => max (ident n) x) 0%nat).
  assert (forall x, In x l -> (ident x <= maxid l)%nat).
  - generalize l. induction l0; simpl; intros.
    + elim H.
    + elim H; intros.
      * subst x. auto with arith. 
      * apply le_trans with (maxid l0); auto with arith. 
  - split.
    + red; intro. generalize (H _ H0). apply le_Sn_n.
    + reflexivity.
Qed.

Lemma fresh_name:
  forall (k: name_kind) (l: list name), exists n, ~In n l /\ kind n = k.
Proof.
  intros. exists (find_fresh_name k l). apply find_fresh_name_is_fresh.
Qed.

(** * Types and typing environments *)

(** ** Type expressions *)

(** IN case of STLC, types are either atomic types or constructed 
  using the funtional constructor. Therefore, no variables are
  necessary. 

  The constructor [Tparam] below is orginally for parameters in
  Leroy's contribution to POPLmark Challenge. To make the modification
  as simple as possible we use the same name although we do not
  condsider them as parameters. Their role is in fact that of atomic
  types. *) 

Inductive type: Set :=
  | Tparam: name -> type
  | Arrow: type -> type -> type.

(** ** Typing environments *)

(** Typing environments are standard: lists of (name, type) pairs.
  Bindings are added to the left of the environment using the
  [cons] list operation.  Thus, later bindings come first. *)

Definition typenv := env name type.

(** In STLC, an environment is well-formed if no name is bound twice
  in the environment. Note that there are no type variables, hence
  all types are well-formed. *)

Notation wf_env e := (ok_env (fun _ _ => True) e).

(** The following lemmas prove useful properties of environments of
  the form [e1 ++ (x, p) :: e2], that is, all bindings of [e2],
  followed by a binding of [p] to [x], followed by all bindings of
  [e1]. *) 


(* the image of lookup for the wf_envs is well-defined even if x is not in the domain of the env  *)
Lemma wf_env_wd `{EqDec} {X} {e : env A X} x y :
  wf_env e ->
  x = y ->
  e ^^ x = e ^^ y.
Proof.
  induction 1; simpl; intros; auto.
  repeat case_var; try intuition congruence. 
Qed. 

(** Well-formedness of environments is preserved by removing a binding. *)

Lemma wf_env_strengthen (e1 e2 : typenv) x p :
  wf_env (e2 ++ (x, p) :: e1) -> 
  wf_env (e2 ++ e1).
Proof.
  induction e2; simpl; intros. 
  - inversion H; auto.
  - inversion H. subst. constructor; auto.
    rewrite dom_append. 
    rewrite dom_append in H3. simpl in H3. destruct_notIn; auto with datatypes.
Qed.

(** Well-formed environments of the form [e2 ++ (x, p) :: e1]
  are such that [x] is not in the domain of [e1] and not free
  in the types listed in [e1]. *)

(** * Syntax and syntactic operations *)

(** We now define the syntax of STLC terms and basic syntactic notions 
  such as free variables, substitutions, and well-formedness of
  terms.

  The syntax of terms is defined as follows. Bound term variables
  are represented by de Bruijn indices, while parameters(free
  variables) are represented by names. *)

Inductive term : Set :=
  | Param: name -> term
  | Var: nat -> term
  | Fun: type -> term -> term
  | App: term -> term -> term.

(** The free names of a term include term names. *)

Fixpoint fv_term (a: term) : list name :=
  match a with
  | Param v => v :: nil
  | Var n => nil
  | Fun t a1 => fv_term a1
  | App a1 a2 => fv_term a1 ++ fv_term a2
  end.

(** We use simultaneous substitions instead of single substitutions. 
  One big advantage of using simultaneous substitutions is that the
  number of definitions and lemmas decreases. 

  We need special environments for simultanous substitutions which we
  call _associations_. *)

Definition v_assoc := env nat term.
Definition p_assoc := env name term.

(** Prerequisites for dealing with de Bruijn indices *)

Fixpoint shifting (e: v_assoc) : v_assoc :=
  match e with
  | nil => nil
  | (x, t) :: e0 => (S x, t) :: shifting e0
  end.

Fixpoint less_no (e: v_assoc) n : nat :=
  match e with
  | nil => 0
  | (i, t) :: e0 => if lt_dec i n
                    then S (less_no e0 n)
                    else less_no e0 n
  end.

(** Simultaneous substitutio for variables and parameters*)

Fixpoint subst_term (a: term) (ep: p_assoc) (ev : v_assoc) {struct a} : term :=
  match a with
  | Param v   => match ep ^^ v with
                 | Some t => t
                 | None => Param v
                 end
  | Var n     => match ev ^^ n with
                 | Some t => t
                 | None => Var (n - (less_no ev n))
                 end
  | Fun t a1  => Fun t (subst_term a1 ep (shifting ev)) 
  | App a1 a2 => App (subst_term a1 ep ev) (subst_term a2 ep ev)
  end.

Notation "a [! ep , ev !]" := (subst_term a ep ev) (at level 69).

(* vsubst_term *)
Notation "a [[! ev !]]" := (subst_term a nil ev) (at level 69).

(* psubst_term *)
Notation "a [[$ ep $]]" := (subst_term a ep nil) (at level 69).


Fixpoint rename_term (r : renaming name) a :=
  match a with
  | Param x => match r ^^ x with
               | Some y => Param y
               | _      => Param x
               end
  | Var n   =>  Var n
  | Fun t a1 => Fun t (rename_term r a1)
  | App a1 a2 => App (rename_term r a1) (rename_term r a2)
  end.

Notation " r $ a " := (@rename_term r a) (at level 60).

Lemma rename_term_nil (a : term) :
  nil $ a = a.
Proof. 
  induction a; simpl; f_equal; auto. 
Qed.


(** Here is the ``freshening'' operations replacing the bound variable
  0 with a term. *)

Definition freshen_term (a: term) (x: name) : term :=
  a [[! (0, Param x) :: nil !]].

(** Substitutions and freshening play well with free variables. *)

Lemma fv_term_vsubst_term x a ev :
  x @ (fv_term a) ->
  x @ (fv_term (a [[! ev !]])).
Proof.
  generalize dependent ev.
  induction a; simpl; intros; intuition.
  elim (in_app_or _ _ _ H); eauto with datatypes. 
Qed. 

Lemma fv_term_freshen_term x a y :
  x @ (fv_term a) ->
  x @ (fv_term (freshen_term a y)).
Proof.
  intros; unfold freshen_term; apply fv_term_vsubst_term; auto.
Qed.

(** ** Well-formedness of terms *)

(** A term is well-formed in a typing environment if:
  - all names [n] appearing free in a [Param n] subterm are of kind [TERM]
    and are bound in the environment;
  - it does not contain free de Bruijn variables.
*)

Inductive wf_term: typenv -> term -> Prop :=
  | wf_term_param: forall (e : typenv) (x : name),
      x @ (dom e) -> wf_term e (Param x)
  | wf_term_fun: forall (e : typenv) (t : type) (a : term) (x : name),    
      x # (fv_term a) -> x # (dom e) ->
      wf_term ((x, t) :: e) (freshen_term a x) -> 
      wf_term e (Fun t a)
  | wf_term_app: forall e a1 a2,
      wf_term e a1 -> wf_term e a2 -> wf_term e (App a1 a2).

Lemma fv_wf_term x e t :
  wf_term e t -> In x (fv_term t) -> x @ (dom e).
Proof.
  induction 1; simpl; intros.
  - intuition congruence.
  - elim IHwf_term; intros; auto using fv_term_freshen_term.
    subst; elim H; auto using in_or_app.
  - elim (in_app_or _ _ _ H1); auto.
Qed.

(** A term well-formed in [e] remains well-formed if extra bindings
  are added to [e]. *) 

Lemma freshen_rename_term_1 x a x' r r0:
  x # fv_term a -> x' # fv_term a -> x' # dom r ->
  r0 = ((x, x') :: (x', rename_name r x) :: r ^- x) ->
  forall m, (r $ a ) [[! (m, Param x') :: nil !]] = r0 $( a [[! (m, Param x) :: nil !]]).
Proof.
  induction a; auto; simpl; intros.
  - assert  (r ^^ n = r0 ^^n); subst r0.
    + case (x == n); intros.
      * elim H; auto. 
      * do 2 (rewrite <- lookup_neq; auto). auto using env_rm_neq. 
    + simpl. case eq_name; intros. 
      * elim H; auto. 
      * case eq_name; intros;
      [ elim H0; auto
      | rewrite <- env_rm_neq; auto; case_env; auto]. 
  - case_dec eq_nat_dec; auto. simpl; rewrite H2; subst. rewrite lookup_eq; auto. 
  - f_equal; auto. 
  - f_equal; auto with datatypes.
Qed.

Lemma freshen_rename_term_2 x a x' r r0:
  x # fv_term a -> x # dom r -> x' # fv_term a -> x' # dom r ->
  r0 = ((x, x') :: r ) ->
  forall m, (r $ a ) [[! (m, Param x') :: nil !]] = r0 $( a [[! (m, Param x) :: nil !]]).
Proof.
  induction a; simpl; intros; auto. 
  - assert  (r ^^ n = r0 ^^n); subst r0.
    + case (x == n); intros.
      * elim H; auto. 
      * rewrite <- lookup_neq; auto using env_rm_neq. 
    + simpl. case eq_name; intros. 
      * elim H; auto. 
      * case_Env; auto. 
  - case_dec eq_nat_dec; auto. simpl; rewrite H3; subst. rewrite lookup_eq; auto. 
  - f_equal; auto. 
  - f_equal; auto with datatypes.
Qed.


Lemma rename_env_fresh_cons `{D : EqDec} {X} (e : env A X) r x x' : 
  x' # dom e ->
  x' # dom r -> 
  x <> x' ->
  r $$ e =  
  ((x, x') :: (x', rename_name r x) :: r ^- x) $$ (((x, x') :: nil) $$ e).
Proof.
  induction e as [ | []]; auto; simpl; intros. f_equal; auto.
  case (x == a); intros; subst; auto; f_equal.
  + unfold rename_name at 4; erewrite lookup_eq; eauto. 
    erewrite <- rename_name_neq_cons; eauto. 
    unfold rename_name at 2. rewrite lookup_eq; auto. 
  + pattern (rename_name ((x, x') :: nil) a); rewrite <- rename_name_neq_cons; auto.
    change (rename_name nil a) with a. do 2 (rewrite <- rename_name_neq_cons; auto). 
    rewrite <- (@rename_name_neq A  _ X r (fun _ x => x) a x (not_eq_sym n)); auto. 
Qed.

Lemma rename_env_fresh_1 r (e : typenv) x x' r0 e0 t:
  x' # dom (r $$ e) ++ dom r ++ dom e ->
  x <> x' -> 
  e0 = ((x, t) :: ((x, x') :: nil) $$ e) ->
  r0 =( (x, x') :: (x', rename_name r x) :: r ^- x) ->
  (((x', t) :: r $$ e) = r0 $$ e0).    
Proof.
  induction e as [| []]; auto; intros; destruct_notIn.
  - subst; simpl. do 2 f_equal. unfold rename_name. rewrite lookup_eq; auto.
  - rewrite H1; simpl; f_equal. 
    + rewrite H2; unfold rename_name; rewrite lookup_eq; auto.
    + f_equal; auto. 
      * f_equal; rewrite H2.
        destruct (x == n); subst; auto. 
        { unfold rename_name at 4; rewrite lookup_eq; auto. 
          rewrite <- rename_name_neq_cons; eauto. 
          unfold rename_name at 2; rewrite lookup_eq; auto. }
        { pattern ( (rename_name ((x, x') :: nil) n)). 
          rewrite <- rename_name_neq_cons; auto. 
          simpl in H5; do 2 (rewrite <- rename_name_neq_cons; auto). 
          rewrite <- (@rename_name_neq name  _ type r 
                                       (fun _ x => x) n x (not_eq_sym n0)); auto. }
      * rewrite H2. apply rename_env_fresh_cons ; simpl in H5; auto. 
Defined. 

Lemma rename_env_fresh_2 r (e : typenv) x x' r0 e0 t:
  x # dom e -> x # dom r -> x' # dom (r $$ e) ++ dom r ++ dom e ->
  x <> x' -> 
  e0 = ((x, t) ::  e) ->
  r0 =( (x, x') :: r) ->
  (((x', t) :: r $$ e) = r0 $$ e0).    
Proof.
  induction e as [| []]; auto; intros.
  - subst; simpl. do 2 f_equal. unfold rename_name; rewrite lookup_eq; auto.
  - simpl in H. subst; simpl; do 2 f_equal. 
    + unfold rename_name; rewrite lookup_eq; auto.
    + rewrite <- rename_name_neq_cons; auto.
    + rewrite rename_env_notIn; auto.
Qed.

Lemma rename_env_dom_rm_incl (e : typenv) x y :
  x<> y -> 
  incl (dom (e ^- x)) (dom (((x, y) :: nil) $$ e)).
Proof.
  induction e as [| []]; auto; intros. 
  simpl. case_dec eq_name.
  - subst; unfold rename_name; rewrite lookup_eq; auto with datatypes. 
  - rewrite <- rename_name_neq_cons; auto.
    unfold incl in *; intros; simpl in *; intuition.
Qed.


Lemma wf_term_env_incr_gen e a :
  wf_term e a ->
  forall e' r, incl (dom e) (dom e') ->
               wf_term (r $$ e') (r $ a).
Proof.
  induction 1; simpl; intros.
  - case_Env; constructor.
    + eauto using rename_env_dom.
    + eauto using notIn_rename_env_dom.
  - set (l := ((x:: nil) ++ fv_term (r $ a) ++ 
                         dom (r $$ e') ++ fv_term a ++ dom r ++ dom e')).
    destruct (fresh_name TERM l) as [x' [F T]]; subst l; destruct_notIn.
    apply wf_term_fun with x'; auto.
    set (e0 := (x, t) :: ((x, x') :: nil) $$ e');
      set (r0 := (x, x') :: (x', rename_name r x) :: r ^- x).
    assert ((freshen_term (r $ a) x') = r0 $ freshen_term a x) as X.
    + apply freshen_rename_term_1; auto. 
    + assert (((x', t) :: r $$ e') = r0 $$ e0) as X1.
      *{ apply rename_env_fresh_1 with (x:=x); simpl in *; auto.
         - solve_notIn. 
       }
      *{ rewrite X, X1; apply IHwf_term; auto.
         simpl; subst e0. apply incl_cons_3. 
         assert (Hxy : x <> x').
         - apply notIn_cons_1 in H3; auto. 
         - specialize (rename_env_dom_rm_incl e' Hxy); intros.
           replace (dom e) with (dom (e ^- x)). 
           + assert (incl (dom (e^-x)) (dom (e'^-x))). 
             * rewrite! rm_dom; apply incl_rm_1; auto.
             * apply incl_tran with (m:= (dom (e' ^- x))); auto. 
           + rewrite <- env_rm_notIn_dom; auto.
       }
  - constructor; auto. 
Qed.

Lemma id_env_rename_term `{D : EqDec} (e : renaming name) (a : term) :
  rename_term (id_env e) a = a.
Proof.
  induction a; induction e; simpl; auto.
  - destruct a;  simpl; destruct (eq_name n n0); subst; auto.  
  - rewrite IHa; auto.
  - rewrite IHa; auto. 
  - f_equal; auto. 
  - f_equal; auto. 
Qed.

Lemma wf_term_env_incr e a :
  wf_term e a -> forall e', incl (dom e) (dom e') -> wf_term e' a.
Proof.
  intros. specialize (wf_term_env_incr_gen H e' (id_env (id_env e'))).
  intros. set (r:=id_env e'). set (r0 :=  r).  set (a0 := r0 $ a). 
  rewrite id_renaming_env, id_env_rename_term in H1; eauto. 
  intros; auto using id_env_rename_name. 
Qed.  

(** ** Properties of term substitutions *)

Fixpoint term_vars_below (a: term) (m : nat) {struct a} : Prop :=
  match a with
  | Param x => True
  | Var n => n < m
  | Fun t b => term_vars_below b (S m)
  | App b c => term_vars_below b m /\ term_vars_below c m
  end.

Lemma term_vars_below_vsubst_term a n a' :
  term_vars_below (a [[! (n, a'):: nil !]]) n -> 
  term_vars_below a  (S n).
Proof.
  generalize dependent n.
  induction a; auto; simpl; intros. 
  - case_dec eq_nat_dec.
    + subst; auto with arith.
    + case_dec lt_dec; simpl in *; omega.
  - auto.
  - firstorder.
Qed.

Lemma wf_term_vars_below_0 a e :
    wf_term e a -> term_vars_below a 0.
Proof.
  induction 1; simpl; firstorder.
  destruct (fresh_name TERM (fv_term a ++ dom e)).
  eapply term_vars_below_vsubst_term; eauto.
Qed.

Fixpoint v_assoc_above (e: v_assoc) (n : nat) {struct e} : Prop :=
  match e  with
  | nil => True
  | (m, t) :: e' => n <= m /\ v_assoc_above e' n
  end.

Lemma v_assoc_above_None (e : v_assoc) n m :
  v_assoc_above e n ->
  m < n ->
  e ^^ m = None.
Proof.
  induction e as [ | [ ]]; auto; intros. inversion H. 
  simpl; destruct (n0 == m); subst; auto.
  - contradict H1; omega. 
  - rewrite IHe; auto. destruct (eq_nat_dec m n0); auto. omega. 
Qed.  

Lemma v_assoc_above_less_no (e : v_assoc) n m :
  v_assoc_above e n ->
  m < n ->
  less_no e m = 0.
Proof.
  induction e as [ | [k ?]]; simpl; intros; auto.
  case_dec lt_dec; [omega | tauto]. 
Qed.
    
Lemma vsubst_term_invariant_below a n e :
  term_vars_below a n ->
  v_assoc_above e n ->
  a [[! e !]] = a.
Proof.
  generalize dependent e.
  generalize dependent n.
  induction a; simpl; intros; auto.
  - case_Env.
    + rewrite (v_assoc_above_None e H0 H) in * |- ; intuition congruence.
    + rewrite (v_assoc_above_less_no e H0 H). f_equal; omega. 
  - f_equal.
    apply IHa with (S n); auto. induction e as [ | []]; auto. simpl. inversion H0; split.  
    + omega.
    + apply IHe; auto. 
  - f_equal; firstorder.
Qed.    
    
Lemma vsubst_term_wf_term a ev e :
  wf_term e a -> a [[! ev !]] = a.
Proof.
  intros. apply vsubst_term_invariant_below with (n := 0).
  - apply wf_term_vars_below_0 with e; auto.
  - induction ev as [ | [] ]; auto; simpl; auto. 
    split; auto. omega. 
Qed.

Lemma psubst_vsubst_term :
  forall e a x b n c,
  wf_term e b -> 
  (a [[$ (x, b) :: nil $]]) [[! (n, c [[$ (x, b) :: nil $]]) :: nil !]]
  = (a [[! (n, c) :: nil !]]) [[$ (x, b) :: nil $]].
Proof.
  induction a; intros; simpl.
  - case_dec eq_name; auto.
    apply vsubst_term_wf_term with e; auto.
  - repeat (case_dec eq_nat_dec; case_dec lt_dec; case_dec eq_nat_dec);
    intuition; try omega; case_dec lt_dec; intuition; try omega.
    simpl; f_equal; omega.
  - f_equal; auto.
  - f_equal; auto.
Qed.

Lemma psubst_freshen_term:
  forall e a x b y,
    wf_term e b ->
    x <> y ->
    freshen_term (a [[$ (x, b) :: nil $]]) y =
    (freshen_term a y) [[$ (x, b) :: nil $]].
Proof.
  intros; unfold freshen_term. 
  rewrite <- (@psubst_vsubst_term e); auto.
  simpl. rewrite eq_name_false; auto.
Qed.

Lemma vsubst_psubst_term:
  forall x a2 a1 n,
    x # fv_term a1 ->
    a1 [[! (n, a2) :: nil !]] =
    (a1 [[! (n, Param x) :: nil !]]) [[$ (x, a2) :: nil $]].
Proof.
  induction a1; simpl; intros.
  - rewrite eq_name_false; tauto. 
  - case_dec eq_nat_dec; subst; auto. 
    + unfold subst_term. simpl. case_dec eq_name; subst; tauto. 
    + case_dec lt_dec; auto; simpl; auto with arith. 
  - f_equal; auto. 
  - f_equal; auto with datatypes.
Qed.

Lemma vsubst_psubst_freshen_term:
  forall x a1 a2,
    x # fv_term a1 ->
    a1 [[! (0, a2) :: nil !]] =
    (freshen_term a1 x) [[$ (x, a2) :: nil $]].
Proof.
  intros; unfold freshen_term. 
  rewrite vsubst_psubst_term with (x:=x); auto. 
Qed.

Lemma strong_env_ok {A X : Type} {l : env A X}:
  strong_env l -> 
  ok_env (fun _ _ => True) l.
Proof.
  induction 1; constructor; auto.
Qed.

(** * Typing rules *)

(** We now define the typing judgement ``term [a] has type [t] in 
  environment [e]'' as an inductive predicate [has_type e a t]. *)

Inductive has_type: typenv -> term -> type -> Prop :=
  | t_var: forall e x t,
      wf_env e -> lookup x e = Some t ->
      has_type e (Param x) t
  | t_abs: forall e t1 a t2 x,
      x # (fv_term a) -> x # (dom e) ->
      has_type ((x, t1) :: e) (freshen_term a x) t2 ->
      has_type e (Fun t1 a) (Arrow t1 t2)
  | t_app: forall e a b t1 t2,
      has_type e a (Arrow t1 t2) -> has_type e b t1 ->
      has_type e (App a b) t2.

(** Well-formedness properties: if [has_type e a t] holds, then
  [e] is a well-formed environment and [a] a well-formed term. *)

Lemma has_type_wf_env e a t :
  has_type e a t -> wf_env e.
Proof.
  induction 1; auto.
  inversion IHhas_type; auto.
Qed.

Lemma has_type_wf_term e a t :
  has_type e a t -> wf_term e a.
Proof.
  induction 1; auto; intros.
  - constructor; eauto using lookup_inv.
  - apply wf_term_fun with x; auto.
  - constructor; auto.
Qed.

Hint Resolve has_type_wf_env has_type_wf_term.

(** The [has_type] predicate is stable by addition of hypotheses. *)

Lemma wf_term_weaken e e' a :
  wf_term e a -> env_weaken e e' -> wf_term e' a.
Proof.
  intros.
  apply wf_term_env_incr with e; auto.
  apply env_weaken_incl_dom; auto.
Qed.

Hint Resolve wf_term_weaken.


(** Well-formed renaming *)

Lemma notIn_img `{D : EqDec} (r : renaming A) :
    strong_env r ->
    forall x, x @ dom r -> rename_name r x # img (r ^-x). 
Proof.
  induction r as [| []]; auto; intros. 
  destruct (a == x); subst; auto. 
  - unfold rename_name; simpl.
    case (x == x); intro. 
    + inversion H; auto. 
      eapply notIn_sublist; eauto using remove_img_sublist.
    + elim n; auto.
  - case ( x== a); intro; [intuition|].
    rewrite rm_change; auto. 
    change ( img ((a, a0) :: r ^- x)) with (a0 :: img (r ^-x)). 
    unfold rename_name; rewrite <- lookup_neq; auto. inversion H; subst.
    change (match r ^^ x with
              | Some b => b
              | None => x
            end) with (rename_name r x).
    apply notIn_cons_3; [| apply IHr]; simpl in H0; auto; [ | intuition].
    + assert (HH : rename_name ((a, a0) :: r) a = a0).
      * unfold rename_name; rewrite lookup_eq; auto.
      * replace (rename_name r x) with (rename_name ((a, a0) :: r) x). 
        rewrite <- HH at 1. apply strong_renaming_inj; auto; simpl; auto. 
        unfold rename_name; rewrite <- lookup_neq; auto. 
Qed.

Lemma ren_env_dom `{D : EqDec} {X} (e : env A X) : 
  wf_env e ->
  forall x x0, 
    x <> x0 -> x @ dom e ->
    incl (dom (((x, x0) :: nil) $$ e)) (x0 :: dom (e^-x)).
Proof.
  induction e as [| []]; auto; intros. 
  simpl in H1; destruct H1; subst; auto.
  - inversion H; subst; auto. 
    simpl. repeat case_var; unfold rename_name.
    + rewrite lookup_eq; auto; apply incl_cons; auto with datatypes. 
      rewrite rename_env_notIn, rename_env_nil, <- env_rm_notIn_dom; auto with datatypes.
    + rewrite <- lookup_neq; simpl; auto.
      red; simpl; intros. apply or_assoc. apply or_comm. apply or_assoc. 
      destruct H2; auto. left; auto. apply or_comm. 
      eapply IHe; eauto. simpl in H1; inversion H1; auto; elim n; auto.
Qed.

Lemma wf_env_ren_wf_env `{D : EqDec} {X} (e : env A X) :
  wf_env e ->
  forall n m, n <> m -> m # dom e ->
              wf_env (((n, m) :: nil) $$ e).
Proof.
  induction 1; simpl; auto; intros. 
  - constructor.
  - apply not_or in H3; destruct H3. destruct (n == a); subst; auto. 
    + unfold rename_name; rewrite lookup_eq; auto; constructor; auto.
      rewrite rename_env_notIn, rename_env_nil; auto.
    + rewrite <- rename_name_neq_cons; auto; constructor; auto.
      change (rename_name nil a) with a.
      case (n <-? dom e); intro. 
      * specialize (ren_env_dom H H2 i); intro. 
        specialize (@remove_dom_sublist _ _ _ e n); intro. 
        red; intro F. apply H0; eauto using incl_tran with datatypes.
      * rewrite rename_env_notIn, rename_env_nil; auto. 
Qed.

Lemma img_notIn `{D : EqDec} : 
  forall (r : renaming A), 
    strong_env r -> 
    forall x x0, x0 # img r -> x <> x0 -> x0 <> rename_name r x. 
Proof.
  induction r as [ | []]; auto; intros. 
  destruct (a == x); subst; auto.
  - unfold rename_name; rewrite lookup_eq; auto. simpl in H0; auto. 
  - rewrite <- rename_name_neq_cons; auto; apply IHr; auto.
    * inversion H; auto. 
    * simpl in H0; auto.  
Qed. 

Lemma strong_env_wf_env `{D : EqDec} {X} {e : env A X} :
  strong_env e -> wf_env e.
Proof.
  induction 1; constructor; auto.
Qed. 

Lemma rename_wf_env_dom `{D : EqDec} {X} {r : renaming A} {e : env A X} a b t:
  wf_env e -> strong_env r -> 
  (a , t) @ e -> (a, b) @ r -> (b, t) @ (r $$ e).
Proof.                         
  induction e as [ | [c u] e]; auto; intros. 
  simpl; simpl in H1. destruct H1; subst; auto. 
  - left; f_equal; inversion H1; subst; auto; unfold rename_name.
    rewrite (ok_env_lookup_In (strong_env_wf_env H0)  a b H2); auto. 
  - right; apply IHe; auto; inversion H; auto. 
Qed.
  

Lemma ren_env_lookup_neq `{D : EqDec} {X} (e : env A X) : 
  forall n m a, n <> a -> m <> a -> m # dom e -> 
              (((n, m) :: nil ) $$ e ) ^^ a = e ^^ a.
Proof. 
  induction e as [ | [] ]; auto; intros.
  destruct (n == a); subst; auto. 
  - simpl.
    unfold rename_name; rewrite lookup_eq; auto. destruct (a0 == a); auto.
    + elim H; auto.
    + destruct (a0 == m); auto.
      * elim H0; auto.  
      * simpl in H1; apply IHe; auto.
  - simpl. rewrite <- rename_name_neq_cons; change (rename_name nil a) with a; auto. case_var; auto. 
    apply IHe; simpl in H1; auto. 
Qed. 

Lemma has_type_weaken_gen (e : typenv) (a : term) (t : type) :
  has_type e a t -> 
  forall (e' : typenv),
    wf_env e' ->
    env_weaken e e' ->
    forall (r : renaming name),
      strong_env r ->
      incl (dom e') (dom r) ->
      has_type (r $$ e') (r $ a) t.
Proof.
  induction 1; simpl; intros.
  - case_Env.
    + apply t_var.
      * apply ok_env_renaming_simply; auto.
      * pose proof H2 as X. unfold env_weaken in X. specialize (X x t H0).
        pose (@lookup_inv name _ type). 
        apply lookup_In in H5; apply lookup_In in X.
        assert ( wf_env( r $$ e' )). apply ok_env_renaming_simply; auto.
        apply (@ok_env_lookup_In _ _ _ (r $$ e') (fun _ _ => True) H6).
        pose proof X. apply env_lookup_In in H7. inversion H7.
        assert (e' ^^ x = Some t); auto. 
        assert (Some x0 = Some t). rewrite <- H8, <- H9; apply wf_env_wd; auto. 
        rewrite H10 in H8; clear H10 H9.
        apply rename_wf_env_dom with (a:=x); auto. 
    + eapply env_lookup_notIn in H5; eauto. 
      assert (x @ dom r). eauto using env_weaken. apply env_weaken_incl_dom in H2; eauto with datatypes. contradiction. 
  - set (l := ((x:: nil) ++ fv_term (r $ a) ++ dom (r $$ e') ++ fv_term a ++ dom r ++ img r ++ dom e')). 
    destruct (fresh_name TERM l) as [x0 [F K]].
    subst l; destruct_notIn; apply notIn_cons_1, not_eq_sym in H6. 
    apply t_abs with x0; auto. 
    *{ set (e0 := if (x <-? dom r) then ((x, t1) :: ((x, x0) :: nil) $$ e') 
                  else ((x, t1) :: e'));
       set (r0:= if (x<-? dom r) then ((x, x0) :: (x0, rename_name r x) :: r ^- x)
                 else ((x, x0) :: r)).
       destruct (x <-? dom r). 
       +{ assert ((freshen_term (r $ a) x0) = r0 $ freshen_term a x) as X.
          - apply freshen_rename_term_1; auto. 
          - { assert (((x0, t1) :: r $$ e') = r0 $$ e0) as X1.
              - apply rename_env_fresh_1 with (x:=x); 
                auto with datatypes. 
              - rewrite X, X1; apply IHhas_type; auto. 
                + subst e0; constructor; auto.
                  *{ case (x <-? dom e'); intro HH0.
                     - apply wf_env_ren_wf_env; auto. 
                     - rewrite rename_env_notIn, rename_env_nil; auto. 
                   }
                  * { case (x <-? dom e'); intro HH0.
                      - apply (@notIn_sublist name _ (x0 :: dom (e' ^- x)) x). 
                        + simpl; apply or_not; split; auto using not_In_dom.
                        + apply ren_env_dom; auto. 
                      - rewrite rename_env_notIn, rename_env_nil; auto. 
                    } 
                + subst e0; auto. apply env_weaken_add; auto.
                  { case (x <-? dom e'); intro HH0.
                    - unfold env_weaken; intros.
                      + destruct (x == a0). 
                        * rewrite e0 in *. apply notIn_dom_None in H0. 
                          rewrite H12 in H0; contradict H0; discriminate. 
                        * rewrite ren_env_lookup_neq; auto. 
                          intro F; rewrite <- F in H12; rewrite F in *; apply lookup_inv in H12. 
                          apply H13; apply env_weaken_incl_dom in H3; auto.
                    - rewrite rename_env_notIn, rename_env_nil; auto. 
                  }
                + subst r0; constructor; auto; try constructor; auto using strong_env_rm.
                  * apply (@notIn_sublist _ _ (dom r) x0 H10); auto using remove_dom_sublist.
                  * apply notIn_img; auto.
                  * simpl in *; apply or_not; split; auto using not_In_dom.
                  * simpl; apply or_not; split; auto using not_eq_sym, img_notIn. 
                    apply (@notIn_sublist _ _ (img r) x0); eauto using remove_img_sublist.
                + subst r0 e0; simpl; apply incl_cons_3; auto. 
                  case (x <-? dom e'); intro. 
                  * assert (wf_env r); auto using strong_env_ok.
                    specialize (ren_env_dom H2 H6 i0); intro.
                    assert (X2 : incl (dom (e' ^-x)) (dom (r ^- x))).  
                    rewrite! rm_dom; auto using incl_rm_1. 
                    apply incl_tran with (m:= ( (x0 :: dom (e' ^- x)))); auto with datatypes. 
                  * rewrite rename_env_notIn, rename_env_nil; auto; apply incl_tl. 
                    rewrite (env_rm_notIn_dom x e' n); rewrite! rm_dom; auto using incl_rm_1.             }
        }
       + { assert (x # dom e'); auto.
           assert ((freshen_term (r $ a) x0) = r0 $ freshen_term a x) as X.
           - apply freshen_rename_term_2; auto. 
           - { assert (((x0, t1) :: r $$ e') = r0 $$ e0) as X1.
               - apply rename_env_fresh_2 with (x:=x); solve_notIn. 
               - rewrite X, X1; apply IHhas_type; auto; try subst e0 r0; auto.
                 + constructor; auto.
                 + auto using env_weaken_add.
                 + constructor; auto. 
                 + simpl; auto with datatypes. 
             }
         }
     }
  - apply t_app with t1; auto. 
    Grab Existential Variables.     
    exact x.
Qed. 

Lemma id_ren_term_id `{D : EqDec} (e : env name type) n : 
  (id_env e) $ (Param n) = Param n.
Proof.
  induction e as [| []]; auto.
  simpl; case_dec eq_name; subst; auto. 
Qed.

Lemma id_env_ren_term (e : env name type) e' a :
  e' = (id_env e) -> 
  (e' $ a) = a.
Proof.
  induction a; auto; intros;
  [ rewrite H; auto using id_ren_term_id
  | |]; simpl; f_equal; auto.
Qed.


Lemma has_type_weaken e a t :
  has_type e a t -> 
  forall e',
    wf_env e' -> env_weaken e e' -> has_type e' a t.
Proof.
  intros.
  rewrite <- (id_renaming_env_self e').
  rewrite <- (@id_env_ren_term e' (id_env e') a); auto.
  apply has_type_weaken_gen with e; auto.
  - eauto using ok_env_id.
  - rewrite id_env_dom; auto.
Qed.


Lemma psubst_notIn_fv_term x a b y:
  x # fv_term a -> x # fv_term b ->
  x # fv_term (a [[$ (y, b) :: nil$]]).
Proof. 
  induction a; auto; intros; simpl. 
  - destruct (eq_name n y); subst; auto.
  - simpl in H; destruct_notIn; auto with datatypes.
Qed. 


Lemma has_type_stable_term_subst_aux e1 b s :
  has_type e1 b s ->
  forall e a t, has_type e a t ->
                forall e2 x, e = e2 ++ (x, s) :: e1 ->
                             has_type (e2 ++ e1)(a [[$ (x, b) :: nil $]]) t.
Proof.
  induction 2; auto; simpl; intros; subst e.
  -                             (* var *)
    { destruct (eq_name x x0); subst; auto.
      + erewrite lookup_env_extends in H1; eauto; simpl in H1.
          replace (if eq_name x0 x0 then Some s else (e2 ++ (x0, s) :: e1) ^^ x0) with (Some s) in H1. 
        * replace t with s ; [idtac | congruence]. 
          apply has_type_weaken with (e := e1); 
            eauto using  wf_env_strengthen, env_concat_weaken.
        * destruct (eq_name x0 x0); auto. elim n; auto. 
      + constructor; eauto using wf_env_strengthen.
        rewrite env_neq in H1; auto.
    } 
  -                             (* abs *)
    { apply t_abs with x; eauto; rewrite dom_append in H1; simpl in H1; destruct_notIn.
      + assert ( incl (fv_term b) (dom e1)); 
        [red; intros; 
        apply has_type_wf_term in H; eauto using fv_wf_term
        | ].
        assert (x # fv_term b); auto with datatypes. 
        apply psubst_notIn_fv_term; eauto.
      + rewrite dom_append; auto with datatypes. 
      + rewrite psubst_freshen_term with (e:=e1); eauto with datatypes.
        rewrite app_comm_cons; apply IHhas_type; auto. 
    }
  -                             (* App *)
    econstructor; eauto.
Qed.

Lemma has_type_stable_term_subst e1 x b s e2 a t :
  has_type e1 b s ->
  has_type (e2 ++ (x, s) :: e1) a t ->
  has_type (e2 ++ e1) (subst_term a ((x, b):: nil) nil) t.
Proof.
  intros.
  apply has_type_stable_term_subst_aux with s (e2 ++ (x, s) :: e1); auto.
Qed.

(** * Dynamic semantics *)

(** The dynamic semantics of STLC is specified by a one-step reduction
  relation, in small-step operational style.  We first define values
  (final results of reduction sequences) as a subset of terms. *)

Inductive isvalue: term -> Prop :=
  | isvalue_fun: forall t a,
      isvalue (Fun t a).


(** We first give a Plotkin-style specification of the reduction
  relation: it uses inductive rules [red_appfun], [red_apparg] and
  instead of contexts to describe reductions inside applications.  The
  rule [red_appabs] is the familiar beta-reduction rule for term. *)

Inductive red: term -> term -> Prop :=
| red_appabs: forall t a v, (* red-beta  *)
    isvalue v ->
    red (App (Fun t a) v) (a [[! (0,v)::nil !]])
| red_appfun: forall a a' b,  (* red-app-1 *)
    red a a' -> red (App a b) (App a' b)
| red_apparg: forall v b b',  (* red-app-2 *)
    isvalue v -> red b b' -> red (App v b) (App v b').

(** * Type soundness proof *)

(** Type soundness for STLC is established by proving the standard
  properties of type preservation (also called subject reduction) and
  progress. *) 

(** ** Preservation *)


(** The preservation theorem states that if term [a] reduces to [a'],
  then all typings valid for [a] are also valid for [a'].  It is
  proved by an outer induction on the reduction and an inner induction
  on the typing derivation. *)

Theorem preservation_aux a a' :
  red a a' ->
  forall e a0 t,
    has_type e a0 t -> forall (EQ: a = a0), has_type e a' t.
Proof. 
  induction 1; induction 1; intros; simplify_eq EQ; clear EQ; intros.
  - subst.
    inversion H0_. subst.
    clear IHhas_type2 IHhas_type1 H0_.
    rewrite (vsubst_psubst_freshen_term x).
    + change e with (nil ++ e). 
      apply has_type_stable_term_subst with t1; auto.
    + auto. 
  - clear IHhas_type1 IHhas_type2.
    subst.
    assert (has_type e a' (Arrow t1 t2)).
    + apply (IHred e a0 (Arrow t1 t2)); auto.
    + apply t_app with t1; auto.
  - clear IHhas_type1 IHhas_type2.
    subst.
    apply t_app with t1; auto.
    apply (IHred e b0 t1); auto.
Qed.    

Theorem preservation e a a' t :
  red a a' ->
  has_type e a t ->
  has_type e a' t.
Proof. 
  intros; apply preservation_aux with a a; auto.
Qed.

(** ** Progress *)

(** The following lemma determines the shape of a value from its type.
  Namely, closed values of function types are function abstractions. *)

Lemma canonical_forms e a t :
  has_type e a t ->
  e = nil ->
  isvalue a ->
  match t with 
  | Arrow t1 t2 => exists s, exists b, a = Fun s b
  | _ => False
  end.
Proof.
  induction 1; intros; subst e.
  - inversion H2.
  - exists t1; exists a; auto.
  - inversion H2.
Qed.

(** The progress theorem shows that a term well-typed in the empty
  environment is never ``stuck'': either it is a value, or it can
  reduce. The theorem is proved by a simple induction on the typing
  derivation for the term and a case analysis on whether the subterms
  of the term are values or can reduce further. *)

Theorem progress_aux e a t :
  has_type e a t ->
  e = nil ->
  isvalue a \/ exists a', red a a'.
Proof. 
  induction 1; intros; subst e.
  - discriminate.
  - left; constructor.
  - right.
    destruct (IHhas_type1 (refl_equal _)) as [Va | [a' Ra]].
    + destruct (IHhas_type2 (refl_equal _)) as [Vb | [b' Rb]].
      * generalize (canonical_forms H (refl_equal _) Va).
        intros [s [c EQ]]. subst a. 
        exists (c [[! (0, b) :: nil !]]). constructor. auto.
      * exists (App a b'). constructor; auto.
    + exists (App a' b). constructor; auto.
Qed.

Theorem progress a t :
  has_type nil a t -> isvalue a \/ exists a', red a a'.
Proof. 
  intros. apply progress_aux with nil t; auto.
Qed.


