Set Implicit Arguments.

Require Export Arith.
Require Export List.

Require Import extralibrary.
Require Import environments.
Require Import sublist.
Require Import Decidable.

Inductive ok_list {A} : list A -> Prop :=
| ok_nil  : ok_list nil
| ok_cons : forall l a, ok_list l -> a # l -> ok_list (a :: l).

Inductive ok_env {A X : Type} (p : env A X -> X -> Prop) :
  env A X -> Prop :=
  ok_env_nil  : ok_env p nil
| ok_env_cons : forall (a : A) (x : X) (e : env A X),
    ok_env p e ->
    a # dom e ->
    p e x -> 
    ok_env p ((a, x) :: e).

Inductive strong_env {A X : Type} : env A X -> Prop :=
  strong_env_nil  : strong_env nil
| strong_env_cons : forall (a : A) (x : X) (e : env A X),
    strong_env e ->
    a # dom e ->
    x # img e ->
    strong_env ((a, x) :: e).

Definition inj_env `{D : EqDec A} {B : Type} (l : list (A * B)) : Prop :=
  forall a b, a @ dom l -> b @ dom l -> a <> b -> l ^^ a <> l ^^ b.

Hint Unfold inj_env.

Lemma ok_env_dom {A X} {p} {e : env A X} :
  ok_env p e ->
  ok_list (dom e).
Proof.
  induction e as [| []];
  intros; simpl; constructor; inversion H; auto.
Qed.
    
Lemma strong_env_dom {A X} {e : env A X} :
  strong_env e ->
  ok_list (dom e).
Proof.
  induction e as [| []];
  intros; simpl; constructor; inversion H; auto.
Qed.
    
Lemma strong_env_img {A X} {e : env A X} :
  strong_env e ->
  ok_list (img e).
Proof.
  induction e as [| []];
  intros; simpl; constructor; inversion H; auto.
Qed.

Lemma strong_env_inj `{D :EqDec A} {X} {l : list (A * X)} : 
  strong_env l ->
  inj_env l.
Proof.
  unfold inj_env.
  induction 1; intros; auto.
  simpl; repeat case_var; try congruence.
  - apply not_eq_sym; apply env_dom_notIn_img; auto.
    eauto using In_dom_neq.
  - apply env_dom_notIn_img; auto.
    eauto using In_dom_neq.
  - eauto using In_dom_neq.
Qed.

Lemma notIn_rm `{D :EqDec A} {l : list A} {x y : A} :
  x # l -> x # (remove dec y l).
Proof.
  induction l; simpl; intros; auto; case_var; simpl; intuition.
Qed.

Lemma remove_dom_sublist `{D :EqDec A} {X} {e : list (A * X)} :
  forall x,
    incl (dom (e ^- x)) (dom e).
Proof.
  induction e as [ | []]; simpl; intros; auto.
  case_var; simpl; intuition.   
Qed.

Lemma remove_img_sublist `{D :EqDec A} {X} {e : list (A * X)} :
  forall x,
    incl (img (e ^- x)) (img e).
Proof.
  induction e as [ | []]; simpl; intros; auto.
  case_var; simpl; intuition. 
Qed.

Lemma notIn_rm_1 `{D :EqDec A} {l : list A} {a : A} :
  a # l -> 
  remove dec a l = l.
Proof.
  induction l; simpl; auto.
  case_var; intros; intuition.
  f_equal; auto.
Qed.

Lemma notIn_sublist `{D :EqDec A} {l : list A} {x : A} :
  x # l ->
  forall m, incl m l -> x # m.
Proof.
  induction l; simpl; intros.
  - rewrite (incl_nil H0); auto.
  - destruct (a <-? m); intros.
    + cut (x # (remove dec a m)).
      * eauto using notIn_rm_2.
      * apply IHl; intuition.
    + pose proof (incl_rm_3 H0).
      rewrite notIn_rm_1 in H1; auto.
Qed.

Lemma strong_env_rm `{D :EqDec A} {X} {e : list (A * X)} :
  strong_env e ->
  forall x, strong_env (e ^- x).
Proof.
  induction 1; intros; simpl;
  [ constructor
  | simpl; case_var; auto; constructor; auto;
  [eapply notIn_sublist; auto; eauto using remove_dom_sublist 
    | intro F; elim H1; apply (remove_img_sublist x0 x); auto] ].
Qed.

(** Renamming environments *)

Definition renaming A := env A A.

Definition rename_name `{D : EqDec A} (r : renaming A) (a : A) :=
  match r ^^ a with
  | Some b => b
  | None   => a
  end.


Lemma rename_name_notIn_dom `{D : EqDec A} (r : renaming A) (a : A) :
  a # dom r ->
  rename_name r a = a.
Proof.
  unfold rename_name; intro H.
  rewrite (notIn_dom_None _ _ H); auto.
Qed.

Lemma strong_renaming_inj `{D :EqDec A} {r : renaming A} : 
  strong_env r ->
  forall a b,
    a @ dom r ->
    b @ dom r ->
    a <> b ->
    rename_name r a <> rename_name r b.
Proof.
  unfold rename_name.
  intros Hstrong a b Hdoma Hdomb Hnot.
  destruct (lookup_exists _ _ Hdoma) as [x Hx]; rewrite Hx.
  destruct (lookup_exists _ _ Hdomb) as [y Hy]; rewrite Hy.
  pose (strong_env_inj Hstrong Hdoma Hdomb Hnot) as Hnot2.
  intro Hxy; subst; congruence.
Qed.

Fixpoint rename_env `{D : EqDec A} {X} (r : renaming A) (f : renaming A -> X -> X) (e : env A X) :=
  match e with
  | nil          => nil
  | (x, t) :: e' => (rename_name r x, f r t) :: (rename_env r f e')
  end.

Notation " r $$ e " := (rename_env r (fun _ x => x) e) (at level 60).

Lemma rename_env_nil `{D : EqDec} {X} (e : env A X) :
  nil $$ e = e.
Proof.
  induction e as [ | [] ? IHe]; simpl; auto.
  rewrite IHe; auto.  
Qed.

(** ** Identity Renaming  *)

Fixpoint id_renaming {A} (l : list A) : renaming A :=
  match l with
  | nil     => nil
  | a :: l' => (a, a) :: id_renaming l'
  end.

(** Construction of special id_renaming from an environment *)

Definition id_env {A X} (e : env A X) : renaming A :=
  id_renaming (dom e).

Hint Unfold id_env.

Lemma id_renaming_dom {A} (l : list A) :
  dom (id_renaming l) = l.
Proof.
  induction l; simpl; try f_equal; auto.
Qed.

Lemma id_renaming_img {A} (l : list A) :
  img (id_renaming l) = l.
Proof.
  induction l; simpl; try f_equal; auto.
Qed.

Lemma id_env_dom {A X} (e : env A X) :
  dom (id_env e) = dom e.
Proof.
  apply id_renaming_dom.
Qed.

Lemma id_env_img {A X} (e : env A X) :
  img (id_env e) = dom e.
Proof.
  apply id_renaming_img.
Qed.

Hint Rewrite @id_renaming_dom @id_renaming_img @id_env_dom @id_env_img : datatypes.

Lemma ok_list_id_renaming {A} (l : list A) :
  ok_list l -> 
  strong_env (id_renaming l).
Proof.
  induction l as [| a l']; simpl; constructor; auto;
  inversion H; autorewrite with datatypes; auto.
Qed.

Lemma ok_env_id {A X} {p} {e : env A X} :
  ok_env p e ->
  strong_env (id_env e).
Proof.
  intro Hok.
  eauto using ok_list_id_renaming, ok_env_dom.
Qed.

Lemma id_renaming_env `{D : EqDec A} {X} (r : renaming A) (e : env A X) :
  (forall x, x @ dom r -> rename_name r x = x) ->
  r $$ e = e.
Proof.
  induction e as [| []]; simpl; intros; auto.
  repeat f_equal; auto.
  destruct (a <-? (dom r)); auto.
  auto using rename_name_notIn_dom.
Qed.

Lemma id_env_rename_name `{D : EqDec} {X} (e : env A X) (a : A) :
  a @ dom (id_env e) ->
  rename_name (id_env e) a = a.
Proof.
  induction e as [| []]; simpl; intros; auto.
  unfold rename_name; simpl. case_var; auto.
  elim H; intro; intuition congruence. 
Qed.

Lemma id_renaming_env_self `{D : EqDec} {X} (e : env A X) :
  (id_env e) $$ e = e.
Proof.
  apply id_renaming_env; intros.
  apply id_env_rename_name; auto.
Qed.

(** Some of tautology *)

Lemma or_not : forall A B : Prop, ~ A /\ ~ B -> ~ (A \/ B).
Proof.
  tauto.
Qed.

Lemma strong_env_dom_In `{D : EqDec} {X} (r : renaming A) a (e : env A X) (f : renaming A -> X -> X) :
  strong_env r ->
  a @ dom r ->
  incl (dom e) (dom r) ->
  a # dom e ->
  rename_name r a # dom (rename_env r f e).
Proof.
  induction e as [| []]; simpl; intros.
  - auto.
  - pose (not_or _ _ H2) as Hnot; elim Hnot; intros.
    apply or_not; split.
    + pose (incl_cons_0 H1) as Ha.
      apply (strong_renaming_inj H Ha H0); auto.
    + eapply IHe; eauto using incl_cons_1.  
Qed.

Definition compatible `{D : EqDec} {X : Type} (p : env A X -> X -> Prop)
           (f : renaming A -> X -> X) :=
  forall (r : renaming A) (e : env A X) (x : X),
    p e x -> p (rename_env r f e) (f r x).

Hint Unfold compatible.

Lemma ok_env_renaming `{D : EqDec} {X} {r : renaming A} {p}
      {e : env A X} (f : renaming A -> X -> X) :
  ok_env p e ->
  compatible p f -> 
  strong_env r ->
  incl (dom e) (dom r) ->
  ok_env p (rename_env r f e).
Proof.
  induction 1; simpl; intros.
  - apply ok_env_nil.
  - apply ok_env_cons; eauto.
   + apply IHok_env; auto. 
      eapply incl_cons_1; eauto.
    + pose proof (incl_cons_0 H4) as Ha.
      eauto using strong_env_dom_In, incl_cons_1.
Qed.

Lemma ok_env_renaming_simply `{D : EqDec} {X} {r : renaming A}
      {p : env A X -> X -> Prop} {e : env A X} :
  ok_env p e ->
  (forall e x, p e x) -> 
  strong_env r ->
  incl (dom e) (dom r) ->
  ok_env p (r $$ e).
Proof.
  intros; auto using ok_env_renaming.
Qed.

(** For ok environments, [env_weaken] and [incl] are 
  equivalent. *)

Lemma ok_env_lookup_In `{D : EqDec} {X} {e : env A X} {p}:
  ok_env p e ->
  forall a x, (a, x) @ e -> e ^^ a = Some x.
Proof.
  induction e as [| []]; simpl; intros. 
  - intuition.
  - case_var; inversion H.
    + elim H0; intros; try congruence.
      elim H5.
      eauto using lookup_inv.
    + elim H0; intros; try congruence.
      apply IHe; auto.
Qed.

Lemma ok_env_weaken_incl `{D : EqDec} {X} {e1 e2 : env A X} {p} :
  ok_env p e1 -> 
  ok_env p e2 -> 
  env_weaken e1 e2 ->
  incl e1 e2.
Proof.
  unfold env_weaken, incl; intros. 
  destruct a as [a x].
  apply lookup_In. 
  apply H1.
  eauto using ok_env_lookup_In.
Qed.

Lemma incl_wf_env_weaken `{D : EqDec} {X} {e1 e2 : env A X} {p} :
  ok_env p e2 -> 
  incl e1 e2 ->  
  env_weaken e1 e2.
Proof.
  unfold incl; unfold env_weaken.
  intro Hok2; generalize dependent e1.
  induction Hok2; simpl; intros.
  - elim (H (a, x)).
    auto using lookup_In.
  - case_var; subst; intros.
    + elim (H1 (a, x0)); intros; auto using lookup_In.
      * congruence.
      * pose (env_dom_In H3); congruence.
    + apply (IHHok2 (e1 ^- a)).
      * intros [a1 t1] Ha0.
        { elim (H1 (a1, t1)); auto; intros.
          - pose (env_dom_In Ha0).
            rewrite (rm_dom e1 a) in i.
            pose (In_rm_neq i).
            inversion H3; congruence.
          - apply (env_rm_incl e1 a); auto.
        }
      * rewrite <- env_rm_neq; auto.
Qed.

(** A special case of weakening is the addition of bindings to an
  existing typing environment, provided the resulting environment is
  well-formed. *) 

Lemma env_concat_weaken `{D : EqDec} {X} {e1 e2 : env A X} {p} :
  ok_env p (e1 ++ e2) -> env_weaken e2 (e1 ++ e2).
Proof.
  intro Hok.
  eapply incl_wf_env_weaken; eauto with datatypes.
Qed.

Definition ok_p_env_incr {A X} (p : env A X -> X -> Prop) :=
  forall (e1 e2 : env A X) (x : X),
    p e1 x -> incl (dom e1) (dom e2) -> p e2 x.

Hint Unfold ok_p_env_incr.

Lemma ok_env_extends `{D : EqDec} {X} {e1 e2 : env A X} {p a x y} :
  ok_p_env_incr p -> 
  ok_env p (e1 ++ (a, x) :: e2) ->
  p e2 y ->
  ok_env p (e1 ++ (a, y) :: e2).
Proof.
  induction e1 as [| []]; simpl; intros.
  - inversion H0. constructor; auto.
  - inversion H0; subst. constructor; auto.
    + rewrite (dom_env_extends e1 e2 a y x); auto.
    + unfold ok_p_env_incr in H.
      apply H with (e1 ++ (a, x) :: e2); auto.
      rewrite (dom_env_extends e1 e2 a x y); auto.
Qed.

Lemma lookup_env_extends `{D : EqDec} {X} {e1 e2 : env A X} {p a b x y} :
  ok_env p (e1 ++ (a, y) :: e2) ->
  (e1 ++ (a, x) :: e2) ^^ b =
  if b == a then Some x else (e1 ++ (a, y) :: e2) ^^ b.
Proof.
  induction e1 as [| []]; simpl;
  repeat case_var; intros; (try inversion H); auto.
  inversion H; subst.
  elim H4.
  unfold dom. rewrite (map_append).
  apply in_or_app. right. simpl. tauto.
Qed.

Lemma ok_env_concat_not_free `{D : EqDec} {X} {e1 e2 : env A X} {p a x} :
  ok_env p (e2 ++ (a, x) :: e1) ->
  a # (dom e1).
Proof.
  induction e2; simpl; intros; inversion H; auto.
Qed.

Lemma lookup_env_concat `{D : EqDec} {X} {e1 e2 : env A X} {p a b x y} :
  ok_env p (e2 ++ (a, x) :: e1) ->
  (e2 ++ (a, x) :: e1) ^^ b = Some y ->
  (a = b /\ y = x) \/ (a <> b /\ (e2 ++ e1) ^^ b = Some y).
Proof.
  induction e2 as [| []]; simpl; case_var; intros; try inversion H; auto.
  - left; intuition congruence.
  - right; split; auto.
    inversion H. intro; subst.
    elim H5.
    unfold dom. rewrite (map_append).
    apply in_or_app. right. simpl. tauto.    
Qed.

(** Extra properties to prove various weakening lemmas *)

Fixpoint renaming_names `(r: renaming A) : list A :=
  match r with
  | nil          => nil
  | (x, y) :: r' => x :: y :: renaming_names r'
  end.

Lemma rename_env_dom `{D : EqDec} {X} {r : renaming A} {e : env A X} a b :
  a @ dom e -> r ^^ a = Some b -> b @ dom (r $$ e).
Proof.                         
  induction e as [ | [c t] e]; auto; intros. 
  simpl; inversion H. 
  - left; replace c with a in *; unfold rename_name; rewrite H0; auto. 
  - right; apply IHe; auto.
Qed.

Lemma notIn_rename_env_dom `{D : EqDec} {X} {r : renaming A} {e : env A X} a :
    a @ (dom e) -> r ^^ a = None -> a @ dom (r $$ e).
Proof.
  induction e as [ | [b t] e]; auto; intros.
  simpl; inversion H. 
  - left; replace b with a in *; unfold rename_name; rewrite H0; auto. 
  - right; apply IHe; auto. 
Qed.

Lemma wf_renaming_inj `{EqDec} {X} {e : env A X} x y :
  strong_env e ->
  x @ dom e ->
  y @ dom e ->
  x <> y ->
  e ^^ x <> e ^^ y.
Proof.
  induction 1; simpl; intros.
  - intuition.
  - repeat case_var; try intuition congruence.
    + subst. elim H3; try congruence; intro.
      elim H4; try congruence; intro.
      pose proof (lookup_exists _ _ H7) as Hey.
      destruct Hey as [y1].
      apply not_eq_sym.
      apply env_dom_notIn_img; auto.
    + elim H3; try congruence; intro.
      pose proof (lookup_exists _ _ H6) as Hex.      
      apply env_dom_notIn_img; auto.
Qed.

Lemma rename_name_neq `{D : EqDec} {X}
      (r : env A A) (f : (env A A) -> X -> X) a b :
  a <> b -> rename_name r a = rename_name (r ^- b) a.
Proof.
  intros; unfold rename_name; rewrite <- env_rm_neq; auto. 
Qed.

Lemma rename_rm_notIn  `{D : EqDec} {X} 
      (e : env A X) (r : renaming A) (f : renaming A -> X -> X) x : 
  x # dom e -> r $$ e = (r ^- x) $$ e.
Proof.
  induction e as [| []]; auto; simpl; intros. 
  erewrite <- rename_name_neq; eauto; f_equal; auto.
Qed.

Lemma rename_name_neq_cons `{D : EqDec} a b (r : renaming A) a0 :
  a <> b -> (rename_name r b) = (rename_name ((a, a0) :: r) b).
Proof. 
  intros; unfold rename_name. 
  rewrite (@env_rm_neq _ _ _ b a ((a, a0) :: r) (not_eq_sym H)); simpl; case_var; [|intuition]. 
  rewrite <- env_rm_neq; auto.
Qed. 

Lemma lookup_rename `{D : EqDec} (r : renaming A) (a t : A) :
  r ^^ a = Some t ->  rename_name r a = t.
Proof.
  induction r as [ | []]; intros. discriminate. 
  simpl. unfold rename_name. rewrite H; auto. 
Qed.

Lemma lookup_non_rename `{D : EqDec} (r : renaming A) (a : A) :
  r ^^ a = None -> rename_name r a = a.
Proof.
  induction r as [ | [] ] ; intros; auto. 
  unfold rename_name; rewrite H; auto. 
Qed.

Lemma rename_env_notIn `{D : EqDec} {X} (e : env A X) (r : renaming A) x x0 :
  x # dom e-> ((x, x0) :: r) $$ e = r $$ e.
Proof. 
  induction e as [ | []]; auto; simpl; intros.  
  rewrite IHe; eauto. do 2 f_equal; rewrite <- rename_name_neq_cons; auto.
Qed.



            
